/*
 * ========================================================================
 * (c) Copyright 2021 Micro Focus or one of its affiliates.
 * ========================================================================
 */
// List of view-names that need type differentiation
var overlap = ["userId","collectedUsersOwners","startTime","endTime"]
// Strings prepended to view-name for localization
var view_alias = {"user":"user", "account":"account", "permission":"permission", "permission_to_holders":"holder", "permissions_from_holder":"holder", "permission_hierarchy_to_child":"hierarchy", "permission_hierarchy_to_parent":"hierarchy"}
//These are string attribute names that we don't want showing up on the report
var undesirables = ["Password"];// Made into an array in case more need to be added

/*
Return the name of the fulfillment target
*/
var getTargetName = function (JsonStr) {
    
    //System fulfillment targets initially do not have a name internally until you make a change to it.
    if (!(JSON.parse(JsonStr)["service-identifier"] === undefined) && JSON.parse(JsonStr)["data-source-name"] === undefined){
        return "NOT_INTERNALLY_DEFINED";
    }

    return JSON.parse(JsonStr)["data-source-name"];
}


/*
Get changes to the service parameters between data source configurations
Returns CTE string containing changes to service parameters
*/
var getServiceParamChanges = function (beforeJsonStr, afterJsonStr, dbPlatform) {

    var beforeJSON = JSON.parse(beforeJsonStr);
    var afterJSON = JSON.parse(afterJsonStr);

    var beforeMap = []; // Associative array to store parameter, value pairs for before configuration 
    var afterMap = []; // Associative array to store parameter, value pairs for after configuration 
    
    //Structure for map variables looks like [["Display_name"] -> [value1,value2], ["Display_name2"] -> [value3]]
    //It's like a Python dictionary with keys and values

    // Add parameter, value pairs to beforeMap
    for (i = 0; i < beforeJSON["service-parms"].length; i++) {
        // If service parameter has options or a drop box of options, get value from options.
        if(beforeJSON["service-parms"][i]["options"] != null && !beforeJSON["service-parms"][i]["hidden"]){
            for(j = 0; j < beforeJSON["service-parms"][i]["options"].length; j++){
                //Check each option box value to see if it matches with the actual value of the service param.
                //If it matches, create key/value pair.
                if(beforeJSON["service-parms"][i]["options"][j]["value"] == beforeJSON["service-parms"][i]["value"]){
                    beforeMap[beforeJSON["service-parms"][i]["display-name"]] = beforeJSON["service-parms"][i]["options"][j]["display-name"];
                    break;
                }
            }
            // if option not found add the value with the name added to the front
            if(!(beforeJSON["service-parms"][i]["display-name"] in beforeMap)){
                beforeMap[beforeJSON["service-parms"][i]["display-name"]] = beforeJSON["service-parms"][i]["name"] + "_" + beforeJSON["service-parms"][i]["value"];
            }
        }
        // If there are no options get value
        else{
            // Only add non-hidden parameters
            if(!beforeJSON["service-parms"][i]["hidden"]){
                beforeMap[beforeJSON["service-parms"][i]["display-name"]] = beforeJSON["service-parms"][i]["value"];
            }
        }
    }
    
    //Same as beforeMap
    // Add parameter, value pairs to afterMap
    for (i = 0; i < afterJSON["service-parms"].length; i++) {
        // If service parameter has options get value from options
        if(afterJSON["service-parms"][i]["options"] != null && !afterJSON["service-parms"][i]["hidden"]){
            for(j = 0; j < afterJSON["service-parms"][i]["options"].length; j++){
                if(afterJSON["service-parms"][i]["options"][j]["value"] == afterJSON["service-parms"][i]["value"]){
                    afterMap[afterJSON["service-parms"][i]["display-name"]] = afterJSON["service-parms"][i]["options"][j]["display-name"];
                    break;
                }
            }
            // if option not found add the value with the name added to the front
            if(!(afterJSON["service-parms"][i]["display-name"] in afterMap)){
                afterMap[beforeJSON["service-parms"][i]["display-name"]] = afterJSON["service-parms"][i]["name"] + "_" + afterJSON["service-parms"][i]["value"];
            }
        }
        
        else{ // If there are no options
            if(!afterJSON["service-parms"][i]["hidden"]){ // Only add non-hidden parameters
                afterMap[afterJSON["service-parms"][i]["display-name"]] = afterJSON["service-parms"][i]["value"];
            }    
        }
    }

    // Get differences between the arrays
    var mapDiff = getMapDiff(beforeMap, afterMap, "service-parms");

    return createCTE(mapDiff,"service-parms", "", 0, 0, dbPlatform);
}


/*
Get differences between scrips between data source configurations
Returns CTE string containing changed transformation scripts
*/
var getScriptChanges = function(beforeJsonStr, afterJsonStr, dbPlatform){
    
    var beforeJSON = JSON.parse(beforeJsonStr);
    var afterJSON = JSON.parse(afterJsonStr);

    var beforeMap = []; // Associative array to store name, script pairs for before configuration
    var afterMap = []; // Associative array to store name, script pairs for after configuration

    // Add values to the arrays
    for (i = 0; i < beforeJSON["ecma-scripts"].length; i++) {
        beforeMap[beforeJSON["ecma-scripts"][i]["name"]] = beforeJSON["ecma-scripts"][i]["script"];
    }

    for (i = 0; i < afterJSON["ecma-scripts"].length; i++) {
        afterMap[afterJSON["ecma-scripts"][i]["name"]] = afterJSON["ecma-scripts"][i]["script"];
    }

    // Get differences between the arrays
    var mapDiff = getMapDiff(beforeMap, afterMap, "ecma-scripts");
    
    // Return a CTE with the differences
    return createCTE(mapDiff,"ecma-scripts", "", 1, 0, dbPlatform);
}


/*
This method gets a list of view parameters that have changes. In the UI these are the fields above the collection attributes (e.g Column Delimiter)
These come from  views -> view-parms
Returns CTE string containing parameter changes for all views
*/
var getChangedFulfillmentParams = function (beforeJsonStr, afterJsonStr, dbPlatform) {

    var beforeJSON = JSON.parse(beforeJsonStr);
    var afterJSON = JSON.parse(afterJsonStr);


    var cte = "";

    // Loop through the views in the configuration
    for (i = 0; i < beforeJSON["views"].length; i++) {

        var beforeMap = []; // Associative array to store parameter, value pairs for before configuration 
        var afterMap = []; // Associative array to store parameter, value pairs for after configuration 

        // Add parameter, value pairs to beforeMap
        for (j = 0; j < beforeJSON["views"][i]["schema-map-filter"]["fulfillment-parms"].length; j++) {
            if (!beforeJSON["views"][i]["schema-map-filter"]["fulfillment-parms"][j]["hidden"] && !beforeJSON["views"][i]["schema-map-filter"]["fulfillment-parms"][j]["read-only-mapping"]){
                //Sometimes the value for fulfillment parms can be under "static-value" or "view-name"
                if (beforeJSON["views"][i]["schema-map-filter"]["fulfillment-parms"][j]["static-value"] === undefined){
                    //For some objects, view-name could be an array. It's usually targets that use a search box to find a user or object
                    //and stores a bit of extra information in the json. In this case, we just want the display-name of it if it is an array.
                    //The other check is for if for some reason it's blank so we can avoid a null pointer error.
                    if (!Array.isArray(beforeJSON["views"][i]["schema-map-filter"]["fulfillment-parms"][j]["view-name"])){
                        beforeMap[beforeJSON["views"][i]["schema-map-filter"]["fulfillment-parms"][j]["display-name"]] = beforeJSON["views"][i]["schema-map-filter"]["fulfillment-parms"][j]["view-name"];
                    }
                    else{//Needs a loop because view-name in this case is an array
                        for (k = 0; k < beforeJSON["views"][i]["schema-map-filter"]["fulfillment-parms"][j]["view-name"].length; k++){
                            beforeMap[beforeJSON["views"][i]["schema-map-filter"]["fulfillment-parms"][j]["display-name"]] = beforeJSON["views"][i]["schema-map-filter"]["fulfillment-parms"][j]["view-name"][k]["displayName"];
                        }
                    }
                }
                else{
                    beforeMap[beforeJSON["views"][i]["schema-map-filter"]["fulfillment-parms"][j]["display-name"]] = beforeJSON["views"][i]["schema-map-filter"]["fulfillment-parms"][j]["static-value"];
                }
            }
        }
        
        
        // Add parameter, value pairs to afterMap
        for (j = 0; j < afterJSON["views"][i]["schema-map-filter"]["fulfillment-parms"].length; j++) {
            if (!afterJSON["views"][i]["schema-map-filter"]["fulfillment-parms"][j]["hidden"] && !afterJSON["views"][i]["schema-map-filter"]["fulfillment-parms"][j]["read-only-mapping"]){
                //Sometimes the value for fulfillment parms can be under "static-value" or "view-name"
                if (afterJSON["views"][i]["schema-map-filter"]["fulfillment-parms"][j]["static-value"] === undefined){
                    if (!Array.isArray(afterJSON["views"][i]["schema-map-filter"]["fulfillment-parms"][j]["view-name"])){
                        afterMap[afterJSON["views"][i]["schema-map-filter"]["fulfillment-parms"][j]["display-name"]] = afterJSON["views"][i]["schema-map-filter"]["fulfillment-parms"][j]["view-name"];
                    }
                    else{
                        for (k = 0; k < afterJSON["views"][i]["schema-map-filter"]["fulfillment-parms"][j]["view-name"].length; k++){
                            afterMap[afterJSON["views"][i]["schema-map-filter"]["fulfillment-parms"][j]["display-name"]] = afterJSON["views"][i]["schema-map-filter"]["fulfillment-parms"][j]["view-name"][k]["displayName"];
                        }
                    }
                }
                else{
                    afterMap[afterJSON["views"][i]["schema-map-filter"]["fulfillment-parms"][j]["display-name"]] = afterJSON["views"][i]["schema-map-filter"]["fulfillment-parms"][j]["static-value"];
                }
            }
        }
        

        // Get differences between the arrays
        var mapDiff = getMapDiff(beforeMap, afterMap, "fulfillment-parms");
        
        // Add UNION ALL if the CTE is not empty
        if(cte != ""){
            cte += "UNION ALL\n"
        }

        // Add Section for view in CTE
        cte += createCTE(mapDiff, "fulfillment-parms", beforeJSON["views"][i]["name"], 2, i, dbPlatform);

    }

    // Return CTE containing parameter changes for all views
    return cte;
}


/*
This method gets a list of attributes that have changed for each sub collector(view) (e.g. permission to holder mapping).
These come from  views -> schema-map-filter -> generic-map
Returns CTE string containing generic-map item changes for all views
*/
var getChangedGenMaps = function (beforeJsonStr, afterJsonStr, localizer, dbPlatform) {

    var beforeJSON = JSON.parse(beforeJsonStr);
    var afterJSON = JSON.parse(afterJsonStr);

    var cte = "";

    // Loop through each item in the views list
    for (i = 0; i < beforeJSON["views"].length; i++) {

        var beforeMap = []; // Associative array to store attribute, value pairs for before configuration 
        var afterMap = []; // Associative array to store attribute, value pairs for after configuration 

        // Add attribute, value pairs to beforeMap
        //For some cases, generic-map can be blank. Usually in system fulfillment configurations.
        if (!(beforeJSON["views"][i]["schema-map-filter"]["generic-map"]===undefined)){
            for (j = 0; j < beforeJSON["views"][i]["schema-map-filter"]["generic-map"].length; j++) {
                if(!beforeJSON["views"][i]["schema-map-filter"]["generic-map"][j]["hidden"] && !beforeJSON["views"][i]["schema-map-filter"]["generic-map"][j]["read-only-mapping"]){
                    beforeMap[beforeJSON["views"][i]["schema-map-filter"]["generic-map"][j]["view-name"]] = beforeJSON["views"][i]["schema-map-filter"]["generic-map"][j]["app-name"];
                }
            }
        }

        // Add attribute, value pairs to afterMap
        if (!(afterJSON["views"][i]["schema-map-filter"]["generic-map"]===undefined)){
            for (j = 0; j < afterJSON["views"][i]["schema-map-filter"]["generic-map"].length; j++) {
                if(!afterJSON["views"][i]["schema-map-filter"]["generic-map"][j]["hidden"] && !afterJSON["views"][i]["schema-map-filter"]["generic-map"][j]["read-only-mapping"]){
                    afterMap[afterJSON["views"][i]["schema-map-filter"]["generic-map"][j]["view-name"]] = afterJSON["views"][i]["schema-map-filter"]["generic-map"][j]["app-name"];
                }
            }
        }

        // Get differences between the arrays
        var mapDiff = getMapDiff(beforeMap, afterMap, view_alias[beforeJSON["views"][i]["name"]], localizer);

        // Add UNION ALL if the CTE is not empty
        if(cte != ""){
            cte += "UNION ALL\n";
        }

        // Add Section for view in CTE
        cte += createCTE(mapDiff, "generic-map", beforeJSON["views"][i]["name"], 3, i, dbPlatform);
    }

    // Return CTE containing attribute changes for all views
    return cte;
}



var getChangeItems = function (beforeJsonStr, afterJsonStr, localizer, dbPlatform){
    
    var beforeJSON = JSON.parse(beforeJsonStr);
    var afterJSON = JSON.parse(afterJsonStr);
    
    var beforeMap = [];//Associative arrays to store keys and values.
    var afterMap = [];
    
    //Some don't have a supportedChangeItems element in their json.
    if (!(beforeJSON["supportedChangeItems"] === undefined)){
        var bKeys = Object.keys(beforeJSON["supportedChangeItems"]); //This gets the keys found in the json object
        //Stores the value of they key in the object to the arrays. The arrays will have a different length
        for (i = 0; i < bKeys.length; i++){
            beforeMap[bKeys[i]] = beforeJSON["supportedChangeItems"][bKeys[i]];
        }
    }
    
    if(!(afterJSON["supportedChangeItems"] === undefined)){
        var aKeys = Object.keys(afterJSON["supportedChangeItems"]);
        for(i = 0; i < aKeys.length; i++){
            afterMap[aKeys[i]] = afterJSON["supportedChangeItems"][aKeys[i]];
        }
    }
        var mapDiff = getMapDiff(beforeMap, afterMap, "supported-change-items", localizer); 
    
    
    return createCTE(mapDiff, "supported-change-items", "", 2, 0, dbPlatform);
}

/*
Function to get the difference between two associative arrays
*/
var getMapDiff = function (beforeMap, afterMap, type, localizer) {

    var paramNames = []; // array to store parameter names
    var beforeValues = []; // array to store before values
    var afterValues = []; // array to store after values

    for (var key in beforeMap) {
        if (key in afterMap) {
            // If key is in both beforeMap and afterMap
            if (beforeMap[key] != afterMap[key]) {
                // If values are different add to the arrays
                if (localizer != undefined) {
                    if(overlap.indexOf(key)!=-1){
                        paramNames.push(localizer.getString((type + key).toUpperCase()));
                    } else{
                        paramNames.push(localizer.getString((key).toUpperCase()));
                    }
                }
                else {
                    paramNames.push(key);
                }
                if(type === "ecma-scripts"){
                    // This is to prevent long scripts from causing errors in oracle. Only need to know if the scrips are different.
                    beforeValues.push("script_a");
                    afterValues.push("script_b");
                } else{
                    beforeValues.push(beforeMap[key]);
                    afterValues.push(afterMap[key]);
                }
            }
            // Remove key from beforeMap and afterMap
            delete beforeMap[key];
            delete afterMap[key];
        } else {
            // Keys only in beforeMap
            if (beforeMap[key] != "") {
                if (localizer != undefined) {
                    if(overlap.indexOf(key)!=-1){
                        paramNames.push(localizer.getString((type + key).toUpperCase()));
                    } else{
                        paramNames.push(localizer.getString((key).toUpperCase()));
                    }
                }
                else {
                    paramNames.push(key);
                }
                if(type === "ecma-scripts"){
                    // This is to prevent long scripts from causing errors in oracle.
                    beforeValues.push("script_a");
                } else{
                    beforeValues.push(beforeMap[key]);
                }
                afterValues.push("");
                delete beforeMap[key];
            }
        }
    }
    // Loop through remaining keys in afterMap
    for (var key in afterMap) {
        // Add parameters that are only in afterMap to arrays
        if (afterMap[key] != "") {
            if (localizer != undefined) {
                if(overlap.indexOf(key)!=-1){
                    paramNames.push(localizer.getString((type + key).toUpperCase()));
                } else{
                    paramNames.push(localizer.getString((key).toUpperCase()));
                }
            }
            else {
                paramNames.push(key);
            }
            
            if(type === "ecma-scripts"){
                // This is to prevent long scripts from causing errors in oracle.
                beforeValues.push("");
                afterValues.push("script_b");
            }else if (type === "supported-change-items") {
                beforeValues.push("false");  //In oracle, if the previous value is blank, it won't be read.
                afterValues.push(afterMap[key]);
            }
            else {
                beforeValues.push("");
                afterValues.push(afterMap[key]);
            }
            delete afterMap[key];
        }
    }

    // Return array containing the arrays for names, before values, and after values 
    return [checkQuote(paramNames), checkQuote(beforeValues), checkQuote(afterValues)];
}

//Check if any value has a quote in it and replace it. If it does, then it will replace it with a \". 
//This is for the CTE so no single quotes disrupt the SQL. Some parm values can have single quotes in them.
//This is meant to go in the mapDiff function.
var checkQuote = function(arr){
    
    returnArr = []

    if (!(arr === undefined)){
         for (var key in arr) {
            if(arr[key] === undefined){
                returnArr[arr] = undefined;
            }
            else if (typeof arr[key] == "boolean"){ //Some key values can be booleans. This makes sure they get recorded.
                returnArr[key] = arr[key]
              }
            else{
                if(arr[key].length != 0){
                  for (var j = 0; j < arr[key].length; j++) {
                      arr[key] = arr[key].replace('\'', '\"'); //Replace the quote in the string (.replace only does one at a time)
                      returnArr[key]  = arr[key];
                  }
                }
                else{
                  returnArr[key] = ""
                }
            }
        }
    }

    return returnArr;
}


/*
Function to creates a CTE that jaspersoft can use as the query in a subreport.
The cte contains the columns type, view_name, attr_name, before_value, and after_value.
*/
var createCTE = function(mapDiff, attrType, viewName, typeOrder, viewOrder, dbPlatform){
    var cte = '';

    for(var i = 0; i < mapDiff[0].length; i++){
        if(i > 0){
            // Add 'UNION ALL' if not the first row 
            cte += "UNION ALL\n";
        }
        if (!(undesirables.indexOf(mapDiff[0][i]) > -1)){ // We don't want to record changes made to undesirable values.
            // Column values for the row
            cte += ("--NOT_EMPTY\n"); // Used for jaspersoft print when statement
            cte += ("SELECT\n");
            cte += ("'" + attrType + "' AS type,\n");
            cte += ("'" + viewName + "' AS view_name,\n");
            cte += ("'" + mapDiff[0][i] + "' AS attr_name,\n");
            cte += ("'" + mapDiff[1][i] + "' AS before_value,\n");
            cte += ("'" + mapDiff[2][i] + "' AS after_value,\n");
            cte += (i + " AS row_order,\n");
            cte += (viewOrder + " AS view_order,\n");
            cte += (typeOrder + " AS type_order\n");

            if(dbPlatform === "oracle"){
                // Oracle needs to select from 'DUAL'
                cte += "FROM DUAL\n";
            }
        }
        else{ //If the attribute name equals something we don't want to show up, add empty cte
            cte += ("SELECT\n");
            cte += ("NULL AS type,\n");
            cte += ("NULL AS view_name,\n");
            cte += ("NULL AS attr_name,\n");
            cte += ("NULL AS before_value,\n");
            cte += ("NULL AS after_value,\n");
            cte += ("0 AS row_order,\n");
            cte += ("0 AS view_order,\n");
            cte += ("0 AS type_order\n");

            if(dbPlatform === "oracle"){
                // Oracle needs to select from 'DUAL'
                cte += "FROM DUAL\n";
            }
        
        }
    }

    if(cte === ''){
        // Add row of null values if cte is empty.
        cte += ("SELECT\n");
        cte += ("NULL AS type,\n");
        cte += ("NULL AS view_name,\n");
        cte += ("NULL AS attr_name,\n");
        cte += ("NULL AS before_value,\n");
        cte += ("NULL AS after_value,\n");
        cte += ("0 AS row_order,\n");
        cte += ("0 AS view_order,\n");
        cte += ("0 AS type_order\n");

        if(dbPlatform === "oracle"){
            // Oracle needs to select from 'DUAL'
            cte += "FROM DUAL\n";
        }
    }

    return cte;
}